/***********************************************************************/
/*                                                                     */
/* RIGI Utilities --- Copyright (C) 1996-2000                          */
/*                    The Rigi Group, University of Victoria           */
/*                                                                     */
/* see the file license.terms for usage and copying policies           */
/*                                                                     */
/* Author: Johannes Martin                                             */
/*                                                                     */
/***********************************************************************/
/* rsf2pl.cc 

   Convert RSF file to Gxl program.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "uniqstr.h"
#include "rsffile.h"

enum ArgType {
  ArgNone, ArgError, ArgHelp, Arg4 = 4, ArgNumeric = 8,
};

static int ParseCmdLine(int argc, char **argv, char **domain) {
  int i, j, k, args = 0;
  char c;

  for (i = 1; i < argc; i++) {
    if (*argv[i] != '-') {
      fprintf(stderr, "Invalid argument \"%s\"\n", argv[i]);
      return (ArgError | ArgHelp);
    }
    k = strlen(argv[i]);
    for (j = 1; j < k; j++) {
      c = *(argv[i] + j);
      switch (c) {
      case 'h':
	args |= (int)ArgHelp;
	break;
      case '4':
	args |= (int)Arg4;
	break;
      case 'd':
	j = k;
	*domain = argv[++i];
	break;
      case 'n':
	args |= (int)ArgNumeric;
	break;
      default:
	fprintf(stderr, "Invalid argument \"%s\"\n", argv[i]);
	return (ArgError | ArgHelp);
      }
    }
  }
  return args;
}

static void PrintHelp(char *progName) {
  fprintf(stderr,
	  "Experimental Rigi Standard Format to GXL Converter (rsf2gxl) - " 
	  VERSION "\n");
  fprintf(stderr,
	  "Usage: %s [-h4n] [-d <domain>]\n", progName);
  fprintf(stderr,
	  "\t-h help\n");
  fprintf(stderr,
	  "\t-4 store 4th tuple in 'file' and 'lineno' attributes\n");
  fprintf(stderr,
	  "\t-d use domain <domain> [cparse]\n");
  fprintf(stderr,
	  "\t-n use numeric IDs for nodes, convert RSF IDs to atrributes\n");
}

typedef set<RsfTuple, less<RsfTuple> > Tuples;
typedef set<UniqueString, less<UniqueString> > StringSet;

static void readDomainFile(const char *domain, 
			   const char *domainFileName,
			   Tuples& tuples) {
  FILE *domainFile;
  String fileName;

  if (getenv("RIGI") != NULL) 
    fileName = String(getenv("RIGI")) + "/Rigi/domain/" + 
      domain + "/" + domainFileName;
  else {
    fprintf(stderr, "$RIGI not set, looking for domain files in ./%s.\n", 
	    domain);
    fileName = String("./") + domain + "/" + domainFileName;
  }
  if ((domainFile = fopen(fileName.c_str(), "r")) == 0) {
    fprintf(stderr, "unable to open Rigi domain file %s.", fileName.c_str());
    exit(1);
  }
  if (!readInput(domainFile, tuples, false, true)) {
    fclose(domainFile);
    fprintf(stderr, "error reading Rigi domain file %s.", fileName.c_str());
    exit(1);
  }
  fclose(domainFile);
}

static void writeAtom(const UniqueString& value) {
  static UniqueString unknown("Unknown");
  if ((value == UniqueString::emptyString) || (value == unknown))
    printf("_");
  else
    printf("'%s'", value.chars());
}

static void writeDomainFacts(const Tuples& tuples, 
			     char *predicate, 
			     const char *order) {
  /*
  for (Tuples::iterator iter = tuples.begin(); iter != tuples.end(); iter++) {
    const char *c = order;
    printf("%s(", predicate);
    writeAtom(iter->rsf[*c - '0']);
    while (*++c) {
      printf(", ");
      writeAtom(iter->rsf[*c - '0']);
    }
    printf(").\n");
  }
  */
}

static void convertDomain(const char* domain, StringSet& attributes) {
  Tuples nodeTypes;
  readDomainFile(domain, "Riginode", nodeTypes);
  writeDomainFacts(nodeTypes, "entity", "0");
  Tuples arcTypes;
  readDomainFile(domain, "Rigiarc", arcTypes);
  writeDomainFacts(arcTypes, "relation", "102");
  Tuples attrNames;
  readDomainFile(domain, "Rigiattr", attrNames);
  for (Tuples::iterator iter = attrNames.begin(); iter != attrNames.end(); 
       iter++) {
    if (iter->rsf[2] != UniqueString::emptyString)
      attributes.insert(iter->rsf[2].string());
    else
      if (iter->rsf[1] != UniqueString::emptyString)
	attributes.insert(iter->rsf[1].string());
      else
	attributes.insert(iter->rsf[0].string());
  }
  for (StringSet::iterator iter = attributes.begin(); iter != attributes.end();
       iter++) {
    //    printf("attribute(_, '%s').\n", iter->chars());
  }
}

static bool isAttribute(const StringSet& attributes, const UniqueString& s) {
  return attributes.find(s) != attributes.end();
}

static int printGxlAttribute(const UniqueString& name,
			      const UniqueString& value) {
  return printf("\n  <attr name=\"%s\">\n    <string>%s</string>\n  </attr>", 
		name.chars(), value.chars());
}

static int printGxlAttribute(const UniqueString& name, const int value) {
  return printf("\n  <attr name=\"%s\">\n    <int>%d</int>\n  </attr>", 
		name.chars(), value);
}

static void checkPreserve4(const RsfTuple& tuple, bool preserve4Tuple) {
  if (preserve4Tuple && (tuple.rsf[3].string() != "")) {
    printGxlAttribute("file", tuple.rsf[3]);
    if (tuple.line != 0) 
      printGxlAttribute("lineno", tuple.line);
  }
}

static bool printGxlTextID(const RsfTuple& tuple, 
			   const StringSet& attributes,
			   bool preserve4Tuple) {
  if (tuple.rsf[0] == RsfTuple::typeString) {
    printf("<node id=\"%s\" type=\"%s\">", tuple.rsf[1].chars(), tuple.rsf[2].chars());
    char *name = strdup(tuple.rsf[1].chars());
    char *c;
    if ((c = strchr(name, '^')) != 0) {
      *c = '\0';
      printGxlAttribute("name", name);
    }
    free(name);
    checkPreserve4(tuple, preserve4Tuple);
    return printf("</node>\n") > 0;
  } 
  if (isAttribute(attributes, tuple.rsf[0])) {
    fprintf(stderr, "warning: ignoring RSF atttribute tuple:\n");
    return tuple.print(stderr, preserve4Tuple);
    //    return printGxlAttribute(tuple.rsf[0], tuple.rsf[2]);
  }
  printf("<arc id=%d source=\"%s\" type=\"%s\" dest=\"%s\">", (int) &tuple, 
	 tuple.rsf[1].chars(), tuple.rsf[0].chars(), tuple.rsf[2].chars());
  checkPreserve4(tuple, preserve4Tuple);
  return printf("</arc>\n") > 0;
}

static bool printGxlNumericID(const RsfTuple& tuple, 
			      const StringSet& attributes,
			      bool preserve4Tuple) {
  if (tuple.rsf[0] == RsfTuple::typeString) {
    printf("<node id=%d type=\"%s\">", tuple.rsf[1].hash(), tuple.rsf[2].chars());
    char *name = strdup(tuple.rsf[1].chars());
    char *c;
    if ((c = strchr(name, '^')) != 0) {
      *c = '\0';
      printGxlAttribute("rsfID", c + 1);
    }
    printGxlAttribute("name", name);
    free(name);
    checkPreserve4(tuple, preserve4Tuple);
    return printf("</node>\n") > 0;
  } 
  if (isAttribute(attributes, tuple.rsf[0])) {
    fprintf(stderr, "warning: ignoring RSF atttribute tuple:\n");
    return tuple.print(stderr, preserve4Tuple);
    //    printGxlAttribute(tuple.rsf[0], tuple.rsf[2]);
  }
  printf("<arc id=%d source=%d type=\"%s\" dest=%d>", (int) &tuple, 
	 tuple.rsf[1].hash(), tuple.rsf[0].chars(), tuple.rsf[2].hash());
  checkPreserve4(tuple, preserve4Tuple);
  return printf("</arc>\n") > 0;
}

int main(int argc, char *argv[]) {
  int args = 0;
  char *domain = "cparse";

  if (argc > 1) {
    args = ParseCmdLine(argc, argv, &domain);
    if (args & ArgHelp) {
      PrintHelp(argv[0]);
      return (args & ArgError);
    }
  }

  bool preserve4Tuple = args & Arg4;
  bool numericIDs     = args & ArgNumeric;

  StringSet attributes;
  if (numericIDs) {
    attributes.insert("rsfID");
    attributes.insert("name");
  }
  if (preserve4Tuple) {
    attributes.insert("file");
    attributes.insert("lineno");
  }
  convertDomain(domain, attributes);

  Tuples tuples;
  if (!readInput(tuples, preserve4Tuple))
    return 1;

  Tuples::const_iterator iter = tuples.begin();

  while (iter != tuples.end()) {
    const RsfTuple& rsfTuple = *iter;
    if (numericIDs) {
      if (!printGxlNumericID(rsfTuple, attributes, preserve4Tuple))
	return 1;
    } else {
      if (!printGxlTextID(rsfTuple, attributes, preserve4Tuple))
	return 1;
    }
    if (!preserve4Tuple)
      // eliminate duplicates
      while ((++iter != tuples.end()) &&
	     (rsfTuple.rsf[0] == (*iter).rsf[0]) &&
	     (rsfTuple.rsf[1] == (*iter).rsf[1]) &&
	     (rsfTuple.rsf[2] == (*iter).rsf[2]));
    else 
      ++iter;
  }
}
